/**
 * Read data from files & fifos
 *
 * @package vzlogger
 * @copyright Copyright (c) 2011, The volkszaehler.org project
 * @license http://www.gnu.org/licenses/gpl.txt GNU Public License
 * @author Steffen Vogel <info@steffenvogel.de>
 */
/*
 * This file is part of volkzaehler.org
 *
 * volkzaehler.org is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * volkzaehler.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with volkszaehler.org. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>

#include "meter.h"
#include "protocols/file.h"
#include "options.h"

int meter_init_file(meter_t *mtr, list_t options) {
	meter_handle_file_t *handle = &mtr->handle.file;

	/* path to file which should be read */
	char *path;
	if (options_lookup_string(options, "path", &path) == SUCCESS) {
		handle->path = strdup(path);
	}
	else {
		print(log_error, "Missing path or invalid type", mtr);
		return ERR;
	}

	/* a optional format string for scanf() */
	char *config_format;
	switch (options_lookup_string(options, "format", &config_format)) {
		case SUCCESS: {
			/**
			 * Compiling the provided format string in a format string for scanf
			 * by replacing the following tokens
			 *
			 * "$v" => "%1$f" (value)
			 * "$i" => "%2$ms" (identifier)		(memory gets allocated by sscanf())
			 * "$t" => "%3$f" (timestamp)
			 */

			int config_len = strlen(config_format);
			int scanf_len = config_len + 8; /* adding extra space for longer conversion specification in scanf_format */

			char *scanf_format = malloc(scanf_len); /* the scanf format string */

			int i = 0; /* index in config_format string */
			int j = 0; /* index in scanf_format string */
			while (i <= config_len && j <= scanf_len) {
				switch (config_format[i]) {
					case '$':
						if (i+1 < config_len) { /* introducing a token */
							switch (config_format[i+1]) {
								case 'v': j += sprintf(scanf_format+j, "%%1$f"); break;
								case 'i': j += sprintf(scanf_format+j, "%%2$ms"); break;
								case 't': j += sprintf(scanf_format+j, "%%3$lf"); break;
							}
							i++;
						}
						break;

					case '%':
						scanf_format[j++] = '%'; /* add double %% to escape a conversion identifier */

					default:
						scanf_format[j++] = config_format[i]; /* just copying */
				}

				i++;
			}

			print(log_debug, "Parsed format string \"%s\" => \"%s\"", mtr, config_format, scanf_format);
			handle->format = scanf_format;

			}
			break;

		case ERR_NOT_FOUND:
			handle->format = NULL; /* disable format matching */
			break;

		default:
			print(log_error, "Failed to parse 'format'", mtr);
			return ERR;
	}

	/* should we start each time at the beginning of the file? */
	/* or do we read from a logfile (append) */
	int rewind;
	switch (options_lookup_boolean(options, "rewind", &rewind)) {
		case SUCCESS:
			handle->rewind = rewind;
			break;

		case ERR_INVALID_TYPE:
			print(log_error, "Invalid type for 'rewind'", mtr);
			return ERR;

		case ERR_NOT_FOUND:
			handle->rewind = FALSE; /* do not rewind file by default */
			break;

		default:
			print(log_error, "Failed to parse 'rewind'", mtr);
			return ERR;
	}

	return SUCCESS;
}

void meter_free_file(meter_t *mtr) {
	meter_handle_file_t *handle = &mtr->handle.file;

	free(handle->path);

	if (handle->format != NULL) {
		free(handle->format);
	}
}

int meter_open_file(meter_t *mtr) {
	meter_handle_file_t *handle = &mtr->handle.file;

	handle->fd = fopen(handle->path, "r");

	if (handle->fd == NULL) {
		print(log_error, "fopen(%s): %s", mtr, handle->path, strerror(errno));
		return ERR;
	}

	return SUCCESS;
}

int meter_close_file(meter_t *mtr) {
	meter_handle_file_t *handle = &mtr->handle.file;

	return fclose(handle->fd);
}

size_t meter_read_file(meter_t *mtr, reading_t rds[], size_t n) {
	meter_handle_file_t *handle = &mtr->handle.file;

	// TODO use inotify to block eading until file changes

	char line[256], *endptr;

	/* reset file pointer to beginning of file */
	if (handle->rewind) {
		rewind(handle->fd);
	}

	int i = 0;
	while (fgets(line, 256, handle->fd) && i < n) {
		char *nl;
		if ((nl = strrchr(line, '\n'))) *nl = '\0'; /* remove trailing newline */
		if ((nl = strrchr(line, '\r'))) *nl = '\0';

		if (handle->format) {
			double timestamp;

			/* at least the value has to been read */
			int found = sscanf(line, handle->format, &rds[i].value, &rds[i].identifier.string, &timestamp);
			if (found >= 1) { // TODO free() space allocated for identifier string
				rds[i].time = dtotv(timestamp); /* convert double to timeval */
				i++; /* read successfully */
			}
		}
		else { /* just reading a value per line */
			rds[i].value = strtod(line, &endptr);
			rds[i].identifier.string = NULL;

			gettimeofday(&rds[i].time, NULL);

			if (endptr != line) {
				i++; /* read successfully */
			}
		}
	}

	return i;
}
